import os
from typing import NewType

FilePointer = NewType("FilePointer", str)


def get_file_pointer_from_path(path: str) -> FilePointer:
    """Returns a file pointer from a path string"""
    return FilePointer(path)


def append_paths_to_pointer(pointer: FilePointer, *paths: str) -> FilePointer:
    """Append directories and/or a file name to a specified file pointer.

    Args:
        pointer: `FilePointer` object to add path to
        paths: any number of directory names optionally followed by a file name to append to the
            pointer

    Returns:
        New file pointer to path given by joining given pointer and path names
    """
    return FilePointer(os.path.join(pointer, *paths))


def does_file_or_directory_exist(pointer: FilePointer) -> bool:
    """Checks if a file or directory exists for a given file pointer

    Args:
        pointer: File Pointer to check if file or directory exists at

    Returns:
        True if file or directory at `pointer` exists, False if not
    """
    return os.path.exists(pointer)



"""Tests of catalog functionality"""

import os
import tempfile

import pytest

from hipscat.catalog import Catalog


def test_load_catalog_small_sky(small_sky_dir):
    """Instantiate a catalog with 1 pixel"""
    cat = Catalog(small_sky_dir)

    assert cat.catalog_name == "small_sky"
    assert len(cat.get_pixels()) == 1


def test_load_catalog_small_sky_order1(small_sky_order1_dir):
    """Instantiate a catalog with 4 pixels"""
    cat = Catalog(small_sky_order1_dir)

    assert cat.catalog_name == "small_sky_order1"
    assert len(cat.get_pixels()) == 4


def test_empty_directory():
    """Test loading empty or incomplete data"""
    with tempfile.TemporaryDirectory() as tmp_dir:
        ## Path doesn't exist
        with pytest.raises(FileNotFoundError):
            Catalog(os.path.join("path", "empty"))

        catalog_path = os.path.join(tmp_dir, "empty")
        os.makedirs(catalog_path, exist_ok=True)

        ## Path exists but there's nothing there
        with pytest.raises(FileNotFoundError):
            Catalog(catalog_path)

        ## catalog_info file exists - getting closer
        file_name = os.path.join(catalog_path, "catalog_info.json")
        with open(
            file_name,
            "w",
            encoding="utf-8",
        ) as metadata_file:
            metadata_file.write('{"catalog_name":"empty"}')

        with pytest.raises(FileNotFoundError):
            Catalog(catalog_path)

        ## partition_info file exists - enough to create a catalog
        file_name = os.path.join(catalog_path, "partition_info.csv")
        with open(
            file_name,
            "w",
            encoding="utf-8",
        ) as metadata_file:
            metadata_file.write("foo")

        catalog = Catalog(catalog_path)
        assert catalog.catalog_name == "empty"



import os

import pytest

DATA_DIR_NAME = "data"
SMALL_SKY_DIR_NAME = "small_sky"
SMALL_SKY_ORDER1_DIR_NAME = "small_sky_order1"
TEST_DIR = os.path.dirname(__file__)

# pylint: disable=missing-function-docstring, redefined-outer-name


@pytest.fixture
def test_data_dir():
    return os.path.join(TEST_DIR, DATA_DIR_NAME)


@pytest.fixture
def small_sky_dir(test_data_dir):
    return os.path.join(test_data_dir, SMALL_SKY_DIR_NAME)


@pytest.fixture
def small_sky_order1_dir(test_data_dir):
    return os.path.join(test_data_dir, SMALL_SKY_ORDER1_DIR_NAME)



"""Conatiner class for properties of a catalog that is being created"""

from typing import List

from hipscat.io import file_io


class CatalogParameters:
    """Container class for catalog properties"""

    def __init__(
        self,
        catalog_name=None,
        input_paths: List[str] = None,
        input_format="csv",
        output_path=None,
        highest_healpix_order=10,
        pixel_threshold=1_000_000,
        epoch="J2000",
        ra_column="ra",
        dec_column="dec",
        id_column="id",
    ):
        self.catalog_name = catalog_name
        self.input_format = input_format
        self.input_paths = input_paths
        self.output_path = output_path
        output_path_pointer = file_io.get_file_pointer_from_path(self.output_path)
        self.catalog_base_dir = file_io.append_paths_to_pointer(
            output_path_pointer, self.catalog_name
        )
        self.catalog_path = str(self.catalog_base_dir)
        file_io.make_directory(self.catalog_base_dir, exist_ok=True)

        self.epoch = epoch
        self.ra_column = ra_column
        self.dec_column = dec_column
        self.id_column = id_column

        self.highest_healpix_order = highest_healpix_order
        self.pixel_threshold = pixel_threshold

    def __str__(self):
        formatted_string = (
            f"  catalog_name {self.catalog_name}\n"
            f"  input format {self.input_format}\n"
            f"  num input_paths {len(self.input_paths)}\n"
            f"  epoch {self.epoch}\n"
            f"  ra_column {self.ra_column}\n"
            f"  dec_column {self.dec_column}\n"
            f"  id_column {self.id_column}\n"
            f"  output_path {self.output_path}\n"
            f"  highest_healpix_order {self.highest_healpix_order}\n"
            f"  pixel_threshold {self.pixel_threshold}\n"
        )
        return formatted_string



"""Tests of catalog creation properties"""

import tempfile

from hipscat.catalog.catalog_parameters import CatalogParameters


def test_create_catalog_info():
    """Test that we accurately write out catalog metadata"""
    with tempfile.TemporaryDirectory() as tmp_dir:
        args = CatalogParameters(
            catalog_name="catalog",
            input_paths=["foo"],
            input_format="csv",
            output_path=tmp_dir,
            highest_healpix_order=0,
            ra_column="ra",
            dec_column="dec",
        )

        ## We didn't specify the catalog path - make sure it exists
        assert args.catalog_path

        formatted_string = str(args)
        assert "catalog" in formatted_string
        assert "csv" in formatted_string
        assert tmp_dir in formatted_string



import pytest

from hipscat.catalog import Catalog
from hipscat.catalog.pixel_node import PixelNode
from hipscat.catalog.pixel_node_type import PixelNodeType

# pylint: disable=missing-function-docstring, redefined-outer-name


@pytest.fixture
def root_pixel_node_data():
    return {
        "hp_order": -1,
        "hp_pixel": -1,
        "node_type": PixelNodeType.ROOT,
        "parent": None,
        "children": None,
    }


@pytest.fixture
def root_pixel_node(root_pixel_node_data):
    return PixelNode(**root_pixel_node_data)


@pytest.fixture
def inner_pixel_node_data(root_pixel_node):
    return {
        "hp_order": 0,
        "hp_pixel": 1,
        "node_type": PixelNodeType.INNER,
        "parent": root_pixel_node,
        "children": None,
    }


@pytest.fixture
def inner_pixel_node(inner_pixel_node_data):
    return PixelNode(**inner_pixel_node_data)


@pytest.fixture
def leaf_pixel_node_data(inner_pixel_node):
    return {
        "hp_order": 1,
        "hp_pixel": 12,
        "node_type": PixelNodeType.LEAF,
        "parent": inner_pixel_node,
        "children": None,
    }


@pytest.fixture
def leaf_pixel_node(leaf_pixel_node_data):
    return PixelNode(**leaf_pixel_node_data)


@pytest.fixture
def small_sky_catalog(small_sky_dir):
    return Catalog(small_sky_dir)


@pytest.fixture()
def small_sky_pixels():
    return [
        {"hp_order": 0, "hp_pixel": 11},
    ]


@pytest.fixture
def small_sky_order1_catalog(small_sky_order1_dir):
    return Catalog(small_sky_order1_dir)


@pytest.fixture()
def small_sky_order1_pixels():
    return [
        {"hp_order": 1, "hp_pixel": 44},
        {"hp_order": 1, "hp_pixel": 45},
        {"hp_order": 1, "hp_pixel": 46},
        {"hp_order": 1, "hp_pixel": 47},
    ]



"""Container class to hold per-partition metadata"""

from hipscat.io import paths, file_io


class PartitionInfo:
    """Container class for per-partition info."""

    METADATA_ORDER_COLUMN_NAME = "Norder"
    METADATA_DIR_COLUMN_NAME = "Dir"
    METADATA_PIXEL_COLUMN_NAME = "Npix"

    def __init__(self, catalog_base_dir: file_io.FilePointer) -> None:
        self.catalog_base_dir = catalog_base_dir

        partition_info_pointer = paths.get_partition_info_pointer(self.catalog_base_dir)
        if not file_io.does_file_or_directory_exist(partition_info_pointer):
            raise FileNotFoundError(
                f"No partition info found where expected: {str(partition_info_pointer)}"
            )

        self.data_frame = file_io.load_csv_to_pandas(partition_info_pointer)

    def get_file_names(self):
        """Get file handles for all partition files in the catalog

        Returns:
            one-dimensional array of strings, where each string is a partition file
        """
        file_names = []
        for _, partition in self.data_frame.iterrows():
            file_names.append(
                paths.pixel_catalog_file(
                    self.catalog_base_dir,
                    partition[self.METADATA_ORDER_COLUMN_NAME],
                    partition[self.METADATA_PIXEL_COLUMN_NAME],
                )
            )

        return file_names



"""Catalog data wrappers"""

from .catalog import Catalog
from .catalog_parameters import CatalogParameters
from .partition_info import PartitionInfo
from .pixel_node import PixelNode
from .pixel_node_type import PixelNodeType



"""Container class to hold catalog metadata and partition iteration"""


from hipscat.catalog.partition_info import PartitionInfo
from hipscat.io import file_io, paths


class Catalog:
    """Container class for catalog metadata"""

    def __init__(self, catalog_path: str = None) -> None:
        self.catalog_path = catalog_path
        self.catalog_base_dir = file_io.get_file_pointer_from_path(catalog_path)
        self.metadata_keywords = None

        self.partition_info = None

        self.catalog_name = None

        self._initialize_metadata()

    def _initialize_metadata(self):
        if not file_io.does_file_or_directory_exist(self.catalog_base_dir):
            raise FileNotFoundError(
                f"No directory exists at {str(self.catalog_base_dir)}"
            )
        catalog_info_file = paths.get_catalog_info_pointer(self.catalog_base_dir)
        if not file_io.does_file_or_directory_exist(catalog_info_file):
            raise FileNotFoundError(
                f"No catalog info found where expected: {str(catalog_info_file)}"
            )

        self.metadata_keywords = file_io.load_json_file(catalog_info_file)
        self.catalog_name = self.metadata_keywords["catalog_name"]
        self.partition_info = PartitionInfo(self.catalog_base_dir)

    def get_pixels(self):
        """Get all healpix pixels that are contained in the catalog

        Returns:
            data frame with per-pixel data.

            The data frame contains the following columns:

            - order: order of the destination pixel
            - pixel: pixel number *at the above order*
            - num_objects: the number of rows in the pixel's partition
        """
        return self.partition_info.data_frame



import os

from hipscat.io.file_io import (
    get_file_pointer_from_path,
    does_file_or_directory_exist,
    append_paths_to_pointer,
    FilePointer,
)


def test_get_pointer_from_path(tmp_path):
    tmp_pointer = get_file_pointer_from_path(str(tmp_path))
    assert str(tmp_pointer) == str(tmp_path)


def test_file_or_dir_exist(small_sky_dir):
    small_sky_pointer = get_file_pointer_from_path(small_sky_dir)
    assert does_file_or_directory_exist(small_sky_pointer)
    catalog_info_string = os.path.join(small_sky_dir, "catalog_info.json")
    catalog_info_pointer = get_file_pointer_from_path(catalog_info_string)
    assert does_file_or_directory_exist(catalog_info_pointer)


def test_file_or_dir_exist_false(small_sky_dir):
    small_sky_pointer = get_file_pointer_from_path(small_sky_dir + "incorrect file")
    assert not does_file_or_directory_exist(small_sky_pointer)


def test_append_paths_to_pointer(tmp_path):
    test_paths = ["folder", "file.txt"]
    test_path = os.path.join(tmp_path, *test_paths)
    tmp_pointer = get_file_pointer_from_path(str(tmp_path))
    assert append_paths_to_pointer(tmp_pointer, *test_paths) == test_path



from __future__ import annotations

import pandas as pd

from hipscat.catalog import PartitionInfo, PixelNode, PixelNodeType


class PixelTree:
    """Sparse Quadtree of HEALPix pixels that make up the HiPSCat catalog

    This class stores each node in the tree, with leaf nodes corresponding to pixels with data
    files.

    There are a number of methods in this class which allow for quickly navigating through the
    tree and performing operations to filter the pixels in the catalog.

    Attributes:
        pixels: Nested dictionary of pixel nodes stored in the tree. Indexed by HEALPix
        order then pixel number root_pixel: Root node of the tree. Its children are a subset of the
        12 base HEALPix pixels
    """

    def __init__(self, partition_info_df: pd.DataFrame):
        """Initialises the Tree from the partition info metadata

        Args:
            partition_info_df: Partition Info metadata file loaded into a pandas DataFrame

        Raises:
            ValueError: Incorrectly configured pixel structure in metadata file
        """
        self.root_pixel = PixelNode(-1, -1, PixelNodeType.ROOT, None)
        self.pixels = {-1: {-1: self.root_pixel}}
        self._create_tree(partition_info_df)

    def __len__(self):
        """Gets the number of nodes in the tree, including inner nodes and the root node

        Returns:
            The number of nodes in the tree, including inner nodes and the root node
        """
        pixel_count = 0
        for _, order_pixels in self.pixels.items():
            pixel_count += len(order_pixels)
        return pixel_count

    def contains(self, hp_order: int, hp_pixel: int) -> bool:
        """Check if tree contains a node at a given order and pixel

        Args:
            hp_order: HEALPix order to check
            hp_pixel: HEALPix pixel number to check

        Returns:
            True if the tree contains the pixel, False if not
        """
        return hp_order in self.pixels and hp_pixel in self.pixels[hp_order]

    def get_node(self, hp_order: int, hp_pixel: int) -> PixelNode | None:
        """Get the node at a given order and pixel

        Args:
            hp_order: HEALPix order to get
            hp_pixel: HEALPix pixel number to get

        Returns:
            The PixelNode at the index, or None if a node does not exist
        """
        if self.contains(hp_order, hp_pixel):
            return self.pixels[hp_order][hp_pixel]
        return None

    def _create_tree(self, partition_info_df: pd.DataFrame):
        """Creates the tree by recursively creating parent nodes until reaching the root from
        each leaf pixel

        Validates the pixel structure as it does so by checking for duplicate pixels or pixels
        defined at multiple orders

        Args:
            partition_info_df: Dataframe loaded from the partition_info metadata
        """
        for _, row in partition_info_df.iterrows():
            self._create_node_and_parent_if_not_exist(
                row[PartitionInfo.METADATA_ORDER_COLUMN_NAME],
                row[PartitionInfo.METADATA_PIXEL_COLUMN_NAME],
                PixelNodeType.LEAF,
            )

    def _create_node_and_parent_if_not_exist(
        self, hp_order: int, hp_pixel: int, node_type: PixelNodeType
    ):
        """Creates a node and adds to `self.pixels` in the tree, and recursively creates parent
        node if parent does not exist

        Args:
            hp_order: HEALPix order to create the node at
            hp_pixel: HEALPix pixel to create the node at
            node_type: Node type of the node to create
        """
        if self.contains(hp_order, hp_pixel):
            raise ValueError(
                "Incorrectly configured catalog: catalog contains duplicate pixels"
            )

        if hp_order == 0:
            self._create_node(hp_order, hp_pixel, node_type, self.root_pixel)
            return

        parent_order = hp_order - 1
        parent_pixel = hp_pixel >> 2
        if not self.contains(parent_order, parent_pixel):
            self._create_node_and_parent_if_not_exist(
                parent_order, parent_pixel, PixelNodeType.INNER
            )

        parent = self.pixels[parent_order][parent_pixel]

        if parent.node_type != PixelNodeType.INNER:
            raise ValueError(
                "Incorrectly configured catalog: catalog contains pixels defined at "
                "multiple orders"
            )

        self._create_node(hp_order, hp_pixel, node_type, parent)

    def _create_node(
        self, hp_order: int, hp_pixel: int, node_type: PixelNodeType, parent: PixelNode
    ):
        """Create a node and add to `self.pixels` in the tree

        Args:
            hp_order: HEALPix order to create the node at
            hp_pixel: HEALPix order to create the node at
            node_type: Node type of the node to create
            parent: Parent node of the node to create
        """
        node = PixelNode(hp_order, hp_pixel, node_type, parent)
        if hp_order not in self.pixels:
            self.pixels[hp_order] = {}
        self.pixels[hp_order][hp_pixel] = node



"""Utility functions for writing metadata files"""

import json
from datetime import datetime
from importlib.metadata import version

import numpy as np
import pandas as pd
import pyarrow.dataset as pds

from hipscat.io import file_io, paths


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy integer types"""

    def default(self, o):
        int_object = o
        if isinstance(int_object, (np.int64, np.ulonglong)):
            return int(int_object)
        return o


def write_json_file(metadata_dictionary: dict, file_pointer: file_io.FilePointer):
    """Convert metadata_dictionary to a json string and print to file.

    Args:
        metadata_dictionary (:obj:`dictionary`): a dictionary of key-value pairs
        file_pointer (str): destination for the json file
    """
    dumped_metadata = json.dumps(metadata_dictionary, indent=4, cls=NumpyEncoder)
    file_io.write_string_to_file(file_pointer, dumped_metadata + "\n")


def write_catalog_info(catalog_parameters, histogram: np.ndarray):
    """Write a catalog_info.json file with catalog metadata

    Args:
        catalog_parameters (:obj:`CatalogParameters`): collection of runtime arguments for
            the new catalog
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
    """
    metadata = {}
    metadata["catalog_name"] = catalog_parameters.catalog_name
    metadata["version"] = version("hipscat")
    now = datetime.now()
    metadata["generation_date"] = now.strftime("%Y.%m.%d")
    metadata["epoch"] = catalog_parameters.epoch
    metadata["ra_kw"] = catalog_parameters.ra_column
    metadata["dec_kw"] = catalog_parameters.dec_column
    metadata["id_kw"] = catalog_parameters.id_column
    metadata["total_objects"] = histogram.sum()

    metadata["origin_healpix_order"] = catalog_parameters.highest_healpix_order
    metadata["pixel_threshold"] = catalog_parameters.pixel_threshold

    catalog_info_pointer = paths.get_catalog_info_pointer(
        catalog_parameters.catalog_base_dir
    )
    write_json_file(metadata, catalog_info_pointer)


def write_provenance_info(catalog_parameters, tool_args):
    """Write a provenance_info.json file with all assorted catalog creation metadata

    Args:
        catalog_parameters (:obj:`CatalogParameters`): collection of runtime arguments
            for the new catalog
        tool_args (:obj:`dict`): dictionary of additional arguments provided by the tool creating
            this catalog.
    """
    metadata = {}
    metadata["catalog_name"] = catalog_parameters.catalog_name
    metadata["version"] = version("hipscat")
    now = datetime.now()
    metadata["generation_date"] = now.strftime("%Y.%m.%d")
    metadata["epoch"] = catalog_parameters.epoch
    metadata["ra_kw"] = catalog_parameters.ra_column
    metadata["dec_kw"] = catalog_parameters.dec_column
    metadata["id_kw"] = catalog_parameters.id_column

    metadata["origin_healpix_order"] = catalog_parameters.highest_healpix_order
    metadata["pixel_threshold"] = catalog_parameters.pixel_threshold

    metadata["tool_args"] = tool_args

    metadata_pointer = paths.get_provenance_pointer(catalog_parameters.catalog_base_dir)
    write_json_file(metadata, metadata_pointer)


def write_partition_info(catalog_parameters, destination_pixel_map: dict):
    """Write all partition data to CSV file.

    Args:
        catalog_parameters (:obj:`CatalogParameters`): collection of runtime arguments for
            the new job
        destination_pixel_map (dict): data frame that has as columns:

            - pixel order of destination
            - pixel number of destination
            - sum of rows in destination
            - list of all source pixels at original order
    """
    partition_info_pointer = paths.get_partition_info_pointer(
        catalog_parameters.catalog_base_dir
    )
    data_frame = pd.DataFrame(destination_pixel_map.keys())

    # Set column names and add a directory column.
    data_frame.columns = [
        "Norder",
        "Npix",
        "num_objects",
    ]
    data_frame["Dir"] = int(data_frame["Npix"] / 10_000) * 10_000

    # Reorder the columns to match full path, and force to integer types.
    data_frame = data_frame[
        [
            "Norder",
            "Dir",
            "Npix",
            "num_objects",
        ]
    ].astype(int)

    file_io.write_dataframe_to_csv(data_frame, partition_info_pointer, index=False)


def write_legacy_metadata(
    catalog_parameters, histogram: np.ndarray, pixel_map: np.ndarray
):
    """Write a <catalog_name>_meta.json with the format expected by the prototype catalog.

    Args:
        catalog_parameters (:obj:`CatalogParameters`): collection of runtime arguments for
            the new catalog
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
        pixel_map (:obj:`np.ndarray`): one-dimensional numpy array of integer 3-tuples.
            See :func:`~hipscat.pixel_math.partition_stats.generate_alignment` for more
            details on this format.
    """
    metadata = {}
    metadata["cat_name"] = catalog_parameters.catalog_name
    metadata["ra_kw"] = catalog_parameters.ra_column
    metadata["dec_kw"] = catalog_parameters.dec_column
    metadata["id_kw"] = catalog_parameters.id_column
    metadata["n_sources"] = histogram.sum()
    metadata["pix_threshold"] = catalog_parameters.pixel_threshold
    metadata["urls"] = catalog_parameters.input_paths

    hips_structure = {}
    temp = [i for i in pixel_map if i is not None]
    unique_partitions = np.unique(temp, axis=0)
    for item in unique_partitions:
        order = int(item[0])
        pixel = int(item[1])
        if order not in hips_structure:
            hips_structure[order] = []
        hips_structure[order].append(pixel)

    metadata["hips"] = hips_structure

    metadata_pointer = file_io.append_paths_to_pointer(
        catalog_parameters.catalog_base_dir,
        f"{catalog_parameters.catalog_name}_meta.json",
    )

    write_json_file(metadata, metadata_pointer)


def write_parquet_metadata(catalog_path):
    """Generate parquet metadata, using the already-partitioned parquet files
    for this catalog

    Args:
        catalog_path (str): base path for the catalog
    """

    dataset = pds.dataset(catalog_path, format="parquet", exclude_invalid_files=True)
    metadata_collector = []

    for hips_file in dataset.files:
        hips_file_pointer = file_io.get_file_pointer_from_path(hips_file)
        single_metadata = file_io.read_parquet_metadata(hips_file_pointer)
        metadata_collector.append(single_metadata)

    ## Write out the two metadata files
    catalog_base_dir = file_io.get_file_pointer_from_path(catalog_path)
    metadata_file_pointer = paths.get_parquet_metadata_pointer(catalog_base_dir)
    common_metadata_file_pointer = paths.get_common_metadata_pointer(catalog_base_dir)

    file_io.write_parquet_metadata(
        dataset.schema, metadata_file_pointer, metadata_collector=metadata_collector
    )
    file_io.write_parquet_metadata(dataset.schema, common_metadata_file_pointer)


def write_fits_map(catalog_path, histogram: np.ndarray):
    """Write the object spatial distribution information to a healpix FITS file.

    Args:
        catalog_parameters (:obj:`CatalogParameters`): collection of runtime arguments for
            the new catalog
        histogram (:obj:`np.ndarray`): one-dimensional numpy array of long integers where the
            value at each index corresponds to the number of objects found at the healpix pixel.
    """
    catalog_base_dir = file_io.get_file_pointer_from_path(catalog_path)
    map_file_pointer = paths.get_point_map_file_pointer(catalog_base_dir)
    file_io.write_fits_image(histogram, map_file_pointer)



"""Utilities for reading and writing catalog files"""

from .paths import pixel_catalog_file, pixel_directory
from .file_io import (
    get_file_pointer_from_path,
    FilePointer,
)
from .write_metadata import (
    write_catalog_info,
    write_legacy_metadata,
    write_parquet_metadata,
    write_partition_info,
    write_provenance_info,
)



import pandas as pd
import pytest

from hipscat.catalog import PartitionInfo, PixelNodeType
from hipscat.catalog.pixel_tree import PixelTree


def assert_pixel_tree_has_nodes_in_catalog(tree, catalog):
    """assert tree contains the same nodes as the catalog"""
    assert tree.contains(-1, -1)
    for _, pixel in catalog.get_pixels().iterrows():
        assert tree.contains(
            pixel[PartitionInfo.METADATA_ORDER_COLUMN_NAME],
            pixel[PartitionInfo.METADATA_PIXEL_COLUMN_NAME],
        )


def test_pixel_tree_small_sky(small_sky_catalog, small_sky_pixels):
    """test pixel tree on small sky"""
    pixel_tree = PixelTree(small_sky_catalog.get_pixels())
    assert len(pixel_tree) == len(small_sky_catalog.get_pixels()) + 1
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_catalog)
    small_sky_pixel = pixel_tree.get_node(**small_sky_pixels[0])
    assert small_sky_pixel.parent == pixel_tree.root_pixel
    assert pixel_tree.root_pixel.node_type == PixelNodeType.ROOT


def test_pixel_tree_small_sky_order1(small_sky_order1_catalog, small_sky_order1_pixels):
    """test pixel tree on small sky order1"""
    pixel_tree = PixelTree(small_sky_order1_catalog.get_pixels())
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_order1_catalog)
    first_pixel = pixel_tree.get_node(**small_sky_order1_pixels[0])
    second_pixel = pixel_tree.get_node(**small_sky_order1_pixels[1])
    parent_node = first_pixel.parent
    assert parent_node.node_type == PixelNodeType.INNER
    assert first_pixel.parent == second_pixel.parent
    assert parent_node.parent == pixel_tree.root_pixel


def test_duplicate_pixel_raises_error(small_sky_catalog):
    """test pixel tree raises error with duplicate pixels"""
    partition_info = small_sky_catalog.get_pixels()
    pixel_row = partition_info.iloc[0]
    info_with_duplicate = pd.concat([partition_info, pixel_row.to_frame().T])
    with pytest.raises(ValueError):
        PixelTree(info_with_duplicate)


def test_pixel_duplicated_at_different_order_raises_error(small_sky_catalog):
    """test pixel tree raises error with duplicate pixels at different orders"""
    partition_info = small_sky_catalog.get_pixels()
    pixel_row = partition_info.iloc[0].copy()
    pixel_row[PartitionInfo.METADATA_ORDER_COLUMN_NAME] += 1
    pixel_row[PartitionInfo.METADATA_PIXEL_COLUMN_NAME] = (
        pixel_row[PartitionInfo.METADATA_PIXEL_COLUMN_NAME] << 2
    )
    info_with_duplicate = pd.concat([partition_info, pixel_row.to_frame().T])
    with pytest.raises(ValueError):
        PixelTree(info_with_duplicate)



"""Tests of partition info functionality"""

import os

from hipscat.catalog import PartitionInfo


def test_load_partition_info_small_sky(small_sky_dir):
    """Instantiate the partition info for catalog with 1 pixel"""
    partitions = PartitionInfo(small_sky_dir)

    partition_file_list = partitions.get_file_names()
    assert len(partition_file_list) == 1

    for parquet_file in partition_file_list:
        assert os.path.exists(parquet_file)


def test_load_partition_info_small_sky_order1(small_sky_order1_dir):
    """Instantiate the partition info for catalog with 4 pixels"""
    partitions = PartitionInfo(small_sky_order1_dir)

    partition_file_list = partitions.get_file_names()
    assert len(partition_file_list) == 4

    for parquet_file in partition_file_list:
        assert os.path.exists(parquet_file)



"""
Compute the hipscat ID field

See the README in this directory for more information on the fiddly pixel math.
"""

import healpy as hp
import numpy as np

HIPSCAT_ID_HEALPIX_ORDER = 19


def compute_hipscat_id(ra_values, dec_values):
    """Compute the hipscat ID field.

    This index is defined as a 64-bit integer which has two parts:
        - healpix pixel (at order 19)
        - incrementing counter (within same healpix, for uniqueness)
    |------------------------------------------|-------------------|
    |<-----    healpixel at order 19    ------>|<--   counter   -->|

    This provides us with an increasing index, that will not overlap
    between spatially partitioned data files.

    Args:
        ra_values (list[float]): celestial coordinates, right ascension
        dec_values (list[float]): celestial coordinates, declination
    Returns:
        one-dimensional array of int64s with hipscat IDs for the sky positions.
    Raises:
        ValueError: if the length of the input lists don't match.
    """
    if len(ra_values) != len(dec_values):
        raise ValueError("ra and dec arrays should have the same length")

    ## Construct the bit-shifted healpix segment
    value_count = len(ra_values)
    mapped_pixels = hp.ang2pix(
        2**HIPSCAT_ID_HEALPIX_ORDER, ra_values, dec_values, nest=True, lonlat=True
    )

    ## We sort to put pixels next to each other that will need to be counted.
    ## This simplifies the counter logic, as we can subtract the index where
    ## we first see the pixel value from the current index to get the offset counter.
    sort_index = np.argsort(mapped_pixels)
    mapped_pixels = mapped_pixels[sort_index]

    ## Construct the counter.
    _, unique_indices, unique_inverse = np.unique(
        mapped_pixels, return_inverse=True, return_index=True
    )
    unique_indices = unique_indices.astype(np.uint64)
    boring_number_index = np.arange(value_count, dtype=np.uint64)
    offset_counter = boring_number_index - unique_indices[unique_inverse]

    ## Add counter to shifted pixel, and map back to the original, unsorted, values
    shifted_pixels = mapped_pixels.astype(np.uint64) << (
        64 - (4 + 2 * HIPSCAT_ID_HEALPIX_ORDER)
    )
    shifted_pixels = shifted_pixels + offset_counter

    unsort_index = np.argsort(sort_index)
    return shifted_pixels[unsort_index]


def hipscat_id_to_healpix(ids):
    """Convert some hipscat ids to the healpix pixel at order 19.
    This is just bit-shifting the counter away.

    Args:
        ids (list[int64]): list of well-formatted hipscat ids
    Returns:
        list of order 19 pixels from the hipscat id
    """
    return np.asarray(ids, dtype=np.uint64) >> (64 - (4 + 2 * HIPSCAT_ID_HEALPIX_ORDER))
